/*! \file talloc.h
 * Convenience wrapper.  libosmocore used to ship its own internal copy of
 * talloc, before libtalloc became a standard component on most systems */
#pragma once
#include <talloc.h>
